#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay_basic.h>
#include <avr/eeprom.h>


//#include "iocompat.h"


/*************************
 *     AVR SWR METER     *
 *  (C) OK1DX 2009-12    *
 *   ok1dx@amsat.org     *
 *************************
 version 1.1x
 - bugfix by UY3QW
 version 1.1
 - SWR calculated using peak measurement
 - indication of FWD overload
 version 1.0
 - first release
 
 */

// Xtal 4 MHz
// fuse bits:
// low byte 11111111 = 0xFF - all unprogrammed BODLEVEL=1, BODEN = 1, SUT0,1 = 1, CKSEL0-3 = 1
// high byte 11011001 (default value from manufacturer, it is not programmed by attached SP12 script)

// SWR meter with LCD 2x40 chars, 2x24 chars or 2x20 chars
#define LCDLINELENGTH 40
//#define LCDLINELENGTH 24
//#define LCDLINELENGTH 20

// uncomment following line if you use internal voltage reference (HW mods necessary, too!)
//#define INTERNAL_REF

// LCD interface:
// PB0 - out, Enable
// PB1 - out, R/W
// PB2 - out, RS (control/data)
// PD4 - in, out D4
// PD5 - in, out D5
// PD6 - in, out D6
// PD7 - in, out D7

#define ENABLE PB0
#define RW PB1
#define RS PB2

// Probe interface:
// PC0 - ADC0, probe select
// PC1 - ADC1, forward voltage
// PC2 - ADC2, return voltage

// Programming interface
// PC6 - RESET
// PB3 - MOSI
// PB4 - MISO
// PB5 - SCK

// USART interface
// PD0 - RXD
// PD1 - TXD

// SWITCH
// PD2 - in, pullup 


// spare
// PC3 - in, pullup
// PC4 - in, pullup
// PC5 - in, pullup
// PD3 - in, pullup

#define PORTBPU 0
#define PORTCPU ((1<<PD3)|(1<<PD4)|(1<<PD5))
#define PORTDPU ((1<<PD2)|(1<<PD3))

#define RXBUFSIZE 10
#define TXBUFSIZE 10
#define PEAKMEMSIZE 20
// average factor = something like time constant for averaging. Higher value - slower reaction.
#define AVERAGEFACTOR 1


//constants for TIMER1 / 1562 pulses, increment to overflow
#define T1H 249
#define T1L 230

#define PROBES 8

uint8_t flagreg;
#define RXCOMPLETE 0	//received text line terminated by <LF>
#define ADCCOMPLETE 1	//tx buffer empty
#define REFRESH 2		//command from timer to perform display data refresh
#define SETUPMODE 3		//when set, setup mode is active
#define OVERLOAD 4		// FWD ADC value overload

volatile uint8_t rxcount;	// number of characters in RX buffer
uint8_t rxbuf[RXBUFSIZE];	// receive buffer

volatile uint8_t txcount; 	//number of characters in TX buffer
uint8_t txbuf[TXBUFSIZE];	// transmit buffer

char n2a[5];	// buffer for conversion from number to ASCII. No terminating 0!

uint16_t adcValue[3];	// ADC values; 0 = PROBE, 1 = forward, 2 = reverse
uint16_t fwdmem[PEAKMEMSIZE];	// forward peak voltage shift register
uint16_t revmem[PEAKMEMSIZE];	// reverse peak voltage shift register
uint16_t fwdavg;	// used for forward voltage floating averaging
uint16_t revavg;	// used for reverze voltage floating averaging
uint16_t fwpwr;		// calculated scalled forward power (peak or average) up to 9999
uint16_t revpwr;	// calculated scalled reverse power (peak or average) up to 9999
uint8_t memptr;		// pointer to last sample in peak memory

uint8_t timerctr;

uint8_t probe;	// probe selector (based on ADC channel 0)

uint8_t corrs[PROBES][16] __attribute__((section(".eeprom"))) = {	{0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},{0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},{0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},{0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
																	{0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},{0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},{0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},{14,18,16,15,15,16,18,25,30,28,27,15,14,7,6,0}};
// power scaling. HF power for ADC corrected reading = 1000. Value between 1000 and 9999 (4 digits)
uint16_t scales[PROBES] __attribute__((section(".eeprom"))) = {1000,2000,3000,4000,5000,6000,7000,8000};
// power range. bit0-3 position of decimal point, bit4-7 displayed unit (pW,nW,uW,mW,W,kW,MW)
uint8_t ranges[PROBES] __attribute__((section(".eeprom"))) = {0x52, 0x52, 0x52, 0x52, 0x52, 0x52, 0x52,0x42};
uint8_t welcome[] __attribute__((section(".eeprom"))) = "SWR Mega 1.1x OK1DX";
uint8_t txtF[] __attribute__((section(".eeprom"))) = " F=";
uint8_t txtR[] __attribute__((section(".eeprom"))) = " R=";
uint8_t txtPwr[] __attribute__((section(".eeprom"))) = " Pwr=";
uint8_t txtSWR[] __attribute__((section(".eeprom"))) = "W SWR=";
uint8_t txtUnknown[] __attribute__((section(".eeprom"))) = " ??? ";
uint8_t txtOverload[] __attribute__((section(".eeprom"))) = " !!! ";
uint8_t txtHigh[] __attribute__((section(".eeprom"))) = " > 10";
uint8_t txtP[] __attribute__((section(".eeprom"))) = " P=";

// generates interrupt every 100 msec; initiates DAC sampling, displays bar indicator
// every fifths cycle (500 msec) initiates data processing
// performs no direct action, just sets the flags, all work is done in main loop
ISR (TIMER1_OVF_vect)
{
	timerctr++;
	if (timerctr > 4)
	{
		timerctr = 0;
		flagreg |= (1<<REFRESH);	// set flag for display refresh
	}
	ADCSRA |= 1<<ADSC; // start next ADC conversion
	TCNT1H = T1H;	// prepare next timer value; high byte must be written first
	TCNT1L = T1L;	// 390 pulses, increment to overflow
}


// interrupt handler for ADC. Read ADC0,1,2 and put the values into value array and set ADCOMPLETE flag
// to start next conversion just set ADCSC bit in ADCSRA register.
ISR (ADC_vect)
{
	uint8_t channel = ADMUX & 3;	// get channel, mask bits possible values 0,1,2
	if (channel < 3)
	{
		int val = ADCL;	// ADCL must be read before ADCH!!!
		adcValue[channel] = val + (ADCH<<8);
	}
	if (channel > 1)
	{
#ifdef INTERNAL_REF
		ADMUX = 0xc0;	// prepare for next startup, internal reference
#else
		ADMUX = 0;	// prepare for next startup
#endif
		flagreg |= (1<<ADCCOMPLETE);
	}
	else
	{
#ifdef INTERNAL_REF
		ADMUX = channel + 1 + 0xc0;// next one, internal reference
#else
		ADMUX = channel + 1;// other bits are always zero
#endif
		ADCSRA |= 1<<ADSC; // autostart next conversion
	}
}

// interrupt rutine for USART data sending
// to activate transmittion clear txempty, fill the buffer, set txcount and enable UDRIE interrupt
ISR (USART_UDRE_vect)
{
	if (txcount > 0)
	{
		UDR = txbuf[0];	// pass character to TX data register
		txcount--;
		for(uint8_t i=0; i<txcount; i++)
			txbuf[i] = txbuf[i+1];	// buffer shift
	}
	if (txcount == 0)
	{
		UCSRB &= ~(1<<UDRIE);	// disable further interrupt on empty TX USART
//		flagreg &= ~(1<<TXBUSY);	// USART not busy
	}
}

// interrupt rutine for USART reception
// when complete text line received -> flag RXCOMPLETE is set.
// after processing set rxcount to zero and enable RXC interrupt
ISR (USART_RXC_vect)
{
	while (UCSRA & (1<<RXC))	// while there are data in USART RX
	{
		uint8_t rxc = UDR;
		
		if (flagreg &(1<<RXCOMPLETE))
			continue;	// ignore received data when old not yet processed

		if (rxc == 10)	// <LF> sets flag RX complete, disable further RX
		{
			flagreg |= 1<<RXCOMPLETE;
		}
		else
		{
			if ((rxcount < RXBUFSIZE - 1) && (rxc != 13))	// buffer not full, <CR> ignored
			{
				rxbuf[rxcount] = rxc;
				rxcount++;			
			}
		}
	}
}


// wait till LCD reports busy (no timeout)
void lcdWait(void)
{
//	return;// simulation, to allow debugging
	PORTB = PORTBPU;
	DDRD &= 0xF;	// set upper 4 pins as input
	uint8_t data = 0xFF;
	while(data) // if BF not set
	{
		// RS = 0, RW = 1, E = 2 pulses (for both nibles)
		// requirement: enable cycle >1microsec, enable active >450 nsec
		PORTB = PORTBPU | (1<<RW);
		__asm__ volatile("nop");
		PORTB = PORTBPU | (1<<RW) | (1<<ENABLE);
		 __asm__ volatile("nop");
		 __asm__ volatile("nop");
		data = PIND & 0x80;	// BF masked
		PORTB = PORTBPU | (1<<RW);
		__asm__ volatile("nop");
		PORTB = PORTBPU | (1<<RW) | (1<<ENABLE);
		__asm__ volatile("nop");
		__asm__ volatile("nop");
		PORTB = PORTBPU | (1<<RW);
	}
	PORTB = PORTBPU;
	DDRD |= 0xF0;	// set upper 4 pins as output
}

// internally used by send command and send char; not for direct use!
void lcdSend(uint8_t x)
{
	PORTD = PORTDPU | (x & 0xF0);// upper nimble
	PORTB |= (1<<ENABLE);	// enable pulse
	PORTB |= (1<<ENABLE);	// dummy, just for delay
	PORTB &= ~(1<<ENABLE);
	PORTD = PORTDPU | (x<<4);
	PORTB |= (1<<ENABLE);	// enable pulse
	PORTB |= (1<<ENABLE);	// dummy, just for delay
	PORTB &= ~(1<<ENABLE);
}

// send data to LCD (ASCII code of character to be displayed)
void lcdSendChar(uint8_t cmd)
{
	lcdWait();
	PORTB = PORTBPU | (1<<RS); //RS=1, RW=0, EN=0
	lcdSend(cmd);
}

// prepare command/address for following writing of data (both CGRAM and display data)
// note: add 0x40 to address for GCRAM or 0x80 for display data
// LCD line 1: cmd = 0x80 to 0xA7, line 2: 0xC0 to 0xD7
void lcdSendCmd(uint8_t cmd)
{
	lcdWait();
	PORTB = PORTBPU; //RS=0, RW=0, EN=0
	lcdSend(cmd);
}

// downloads character generator RAM (used for bar indicator)
// address should be set before - lcdSetAddr
void lcdInitCGRAM(uint8_t upper, uint8_t lower)
{
	lcdSendChar(upper);
	lcdSendChar(upper);
	lcdSendChar(upper);
	lcdSendChar(0);	// separator
	lcdSendChar(lower);
	lcdSendChar(lower);
	lcdSendChar(lower);
	lcdSendChar(0);	// separator
}

// conversion of unsigned int (value 0 to 9999) into ASCII string, including decimal point
// decimal = location of '.' symbol. If =0, not used.
// fills always the whole buffer (5 characters).
void n2ascii(uint16_t x, uint8_t decimal)
{
	uint16_t base = 1000;
	uint8_t posn = 0;	// buffer position
	for (uint8_t i=0; i<4; i++)
	{
		uint16_t a = x/base;
		x = x % base;
		base /= 10;
		n2a[posn] = (uint8_t)a + '0';
		posn++;
		if(posn==decimal)
		{
			n2a[posn] = '.';
			posn++;
		}
	}
	if (posn<=4)
		n2a[4]=' ';		// decimal point not used, append space instead
	uint8_t i=0;
	while (i<3)
	{
		if ((n2a[i] != '0') || (n2a[i+1] == '.'))
			break;
		n2a[i] = ' ';	// replace leading zeroes by space
		i++;
	}
}


// converts ascii string into uint (in fact the number can be unsigned 0 to 9999
uint16_t ascii2n(uint8_t* buffer, uint8_t len)
{
	uint8_t* bufptr = buffer;
	uint16_t retval = 0;
	while((*bufptr >='0') && (*bufptr <='9') && ((bufptr-buffer) < len))
	{
		retval = 10*retval + *bufptr - '0';
		bufptr++;
	}
	return retval;
}



// append number in n2a buffer to usart TX buffer. Ignore space characters
void usartNumber(void)
{
	for(uint8_t i=0; i<5; i++)
	{
		if (n2a[i] != ' ')
		{
			txbuf[txcount] = n2a[i];
			txcount++;
			if (txcount >= TXBUFSIZE)
				txcount = TXBUFSIZE-1;	// should never happen!!!
		}
	}
}


//periodical sending ADC values when in SETUP mode
void setupUsartSend(void)
{
	if (!(UCSRB & (1<<UDRIE)))	// no other sending in progress
	{
		txbuf[0] = 'P';	// probe
		txcount=1;
		n2ascii(adcValue[0],0);
		usartNumber();
		txbuf[txcount++] = 13;	//<CR>
		txbuf[txcount++] = 10;	//<LF>
		UCSRB |= (1<<UDRIE);	// send it!!!
		
		while (UCSRB & (1<<UDRIE));	// it is sent completely when UDRIE cleared by interrupt rutine
		txbuf[0] = 'F';	// forward
		txcount=1;
		n2ascii(adcValue[1],0);
		usartNumber();
		txbuf[txcount++] = 13;	//<CR>
		txbuf[txcount++] = 10;	//<LF>
		UCSRB |= (1<<UDRIE);	// send it!!!
		
		while (UCSRB & (1<<UDRIE));
		txbuf[0] = 'R';	// reverse
		txcount=1;
		n2ascii(adcValue[2],0);
		usartNumber();
		txbuf[txcount++] = 13;	//<CR>
		txbuf[txcount++] = 10;	//<LF>
		UCSRB |= (1<<UDRIE);	// send it!!!
	}
}

// sends text from EEPROM to display to currnt position
void lcdEText(uint8_t* addr)
{
	uint8_t c = 0;
	do
	{
		c = eeprom_read_byte(addr);
		addr++;
		if (c)
			lcdSendChar(c);
	}
	while(c);
}

// initialize all hardware - ports, timer, usart, LCD,...
void initHW(void)
{
	// general data direction: DDRxn = 0 -> input, 1 -> output. PORTxn = 0 -> pullup OFF, 1 -> pullup ON
	
	// PORT B: bit 0-2 out, bit 3-5 inp 3state. bit 6,7 XTAL - defined by config bits
	DDRB = (1<<PB0)|(1<<PB1)|(1<<PB2);
	PORTB = PORTBPU;
	
	// PORT C: bit 0-2 ADC inp 3state, bit 3-5 in pullup, bit 6 RESET (config bits) tristate
	DDRC = 0;
	PORTC = PORTCPU;
	
	// PORT D: bit 0 RXD in 3state, bit 1 TXD out, bit 2 in pullup switch, bit 3 inp pullup, bit 4-7 out LCD data
	DDRD = (1<<PD1)|(1<<PD4)|(1<<PD5)|(1<<PD6)|(1<<PD7);
	PORTD = PORTDPU | 0x20;	// LCD in 4bit mode
	
// ADC init: ADCMUX keep on default - using AREF pin as voltage reference, right adjust
	// switch ADC on, allow interrupt, precaller 1:32 (clock 125 kHz, conversion time 0.1 msec)
	ADCSRA = (1<<ADEN)|(1<<ADIE)|(1<<ADPS0)|(1<<ADPS2);
#ifdef INTERNAL_REF
	ADMUX = 0xc0; // use internal voltage reference
#endif
	
// USART INIT
	UBRRH = 0;
	UBRRL = 25;	// 9600 Bdby 4 MHz clock

	// enable RX and TX, enable interrupt on RX and TX buffer empty.
	UCSRB = (1<<RXEN)|(1<<TXEN)|(1<<RXCIE);
	/* Set frame format: 8data, 2stop bit */
	UCSRC = (1<<URSEL)|(1<<USBS)|(3<<UCSZ0);

	rxcount = 0;
	txcount = 0;
	
	for(uint8_t i=0; i<3; i++)
		adcValue[i] = 0;	// clear last ADC reading registers
		
	for(uint8_t i=0; i<PEAKMEMSIZE; i++)
	{
		fwdmem[i] = 0;
		revmem[i] = 0;	// clear peak memory buffer
	}
	
	fwdavg = 0;	// clear floating average registers
	revavg = 0;
	probe = 0;	// probe selector
	flagreg = 0; 	// flag register cleared
	
	timerctr = 0;
	
// LCD INIT

	_delay_loop_2(50000);	// 50 msec for 4 MHz clock
	PORTB |= (1<<ENABLE);
	PORTB |= (1<<ENABLE);	// dummy, just short delay
	PORTB &= ~(1<<ENABLE);
	_delay_loop_2(10000);	// another 10 msec
	PORTB |= (1<<ENABLE);
	PORTB |= (1<<ENABLE);	// dummy, just short delay
	PORTB &= ~(1<<ENABLE);
	_delay_loop_2(200);	// another 200 microsec
	PORTB |= (1<<ENABLE);
	PORTB |= (1<<ENABLE);	// dummy, just short delay
	PORTB &= ~(1<<ENABLE);
	PORTD = PORTDPU | 0x20;	// for 8 bit mode
	PORTB |= (1<<ENABLE);
	PORTB |= (1<<ENABLE);	// dummy, just short delay
	PORTB &= ~(1<<ENABLE);	// since now we can use standard lcdSend
	lcdSendCmd(0x28);	// 4 bit, 2 lines, 5x7 mode
	lcdSendCmd(0x08);	// display OFF
	lcdSendCmd(0x0C);	// display ON, no cursor, no blink
	lcdSendCmd(0x06);	// increment mode, no display shift
	lcdSendCmd(0x01);	// display clear
#if LCDLINELENGTH==40
	lcdSendCmd(0x8A);	// go to 1st line, 10th position
#else
	lcdSendCmd(0x80);	// go to 1st line, 1st position
#endif
	lcdEText(welcome);
	_delay_loop_2(0);
	_delay_loop_2(0);
	_delay_loop_2(0);	// 3x 65 msec = 200 msec

    // init LCD Graphic Character Generator
	lcdSendCmd(0x40);	// set initial address GCRAM address
	// code 0 = fwd half, rev nil
	lcdInitCGRAM(0x18, 0);
	// code 1 = fwd nil, rev half
	lcdInitCGRAM(0, 0x18);
	// code 2 = fwd half, rev half
	lcdInitCGRAM(0x18, 0x18);	
	// code 3 = fwd full, rev full
	lcdInitCGRAM(0x1F, 0x1F);
	// code 4 = fwd full, rev nil
	lcdInitCGRAM(0x1F, 0);	
	// code 5 = fwd half, rev full
	lcdInitCGRAM(0x18, 0x1F);	
	// code 6 = fwd full, rev half
	lcdInitCGRAM(0x1F, 0x18);
	// code 7 = fwd nil, rev full
	lcdInitCGRAM(0, 0x1F);

// TIMER1 INIT
	TCCR1A = 0;
	TCCR1B = (1<<CS12);	// prescaler 1:256, 1 pulse per 64 microsec
	TCNT1H = T1H;	// high byte must be written first
	TCNT1L = T1L;
	TIFR = 0;
	TIMSK = (1<<TOIE1);	// enable interrupt for overflow

	sei();	// global interrupt enable; fun starts here
}


// *** correction of ADC value to probe detector non-linearity ***
// linear interpolation using table in EEPROM. The whole range of X value (0-1023)
// is splitted into 16 parts. The EEPROM table contains correction (+ or -) to be added
// to X value to get compensated value
int16_t correction(uint16_t x, uint8_t probe)
{
	if (x > 0x3FF)
		x = 0x3FF;
	uint8_t* idx = corrs[probe] + (x >> 6);	// x value 10 bits, idx 4 bits
	int8_t a = (int8_t)eeprom_read_byte(idx);	// point above setpoint
	int8_t b =  0;
	if ((x>>6) > 0)
		b = (int8_t)eeprom_read_byte(idx - 1);	// point bellow setpoint
	int8_t diff = x & 0x3F;	// lower 6 bits, range between corr points
	return x + b + (((a-b) * diff)/64);
}

uint16_t correction_old(uint16_t x, uint8_t probe)
{	
	return x; //todo: use EEPROM table to correct linearity of detector
}
	

// Show horizontal level bars for forward/reverse signal on 1st display line
void showBar(uint16_t fwd, uint16_t rev)
{
	lcdSendCmd(0x80);	// address - first line, display text
	uint8_t step = 1024/LCDLINELENGTH;
	fwd *= 2;	// we have 2 levels per character
	rev *= 2;
	uint16_t level = step/2;
	for (uint8_t psn = 0; psn < LCDLINELENGTH; psn++)
	{
		uint8_t c = 0;
		if (fwd > level)
			c=1; 
		if (rev > level)
			c|=2;
		level += step;
		if (fwd > level)
			c|=4;
		if (rev > level)
			c|=8;
		level += step;

		if (c == 15)
			c = 3;
		else if (c == 11)
			c = 5;
		else if (c == 10)
			c = 7;
		else if (c == 0)
			c = 32;	// space
		else
			c--;
		lcdSendChar(c);
	}
}


// print number from n2a buffer to LCD at given position
void lcdNumber(void)
{
	for(uint8_t i=0; i<5; i++)
		lcdSendChar(n2a[i]);
}


// multiple 2 12bit numbers, result dividen by 1024 (total 14 bit length)
uint16_t dxMult(uint16_t a, uint16_t b)
{
	uint16_t y;
	uint8_t a1 = (a>>8);	// max 4 bit
	uint16_t a2 = a & 0xFF;
	uint8_t b1 = (b>>8);	// max 4 bit
	uint16_t b2 = b & 0xFF;
	y = (((a1 * b1)<<8) + (a1 * b2) + (a2 * b1) + ((a2 * b2)>>8))>>2;
	return y;
}

// calculates power based on measured voltage  scalled to range
// range = power when voltage = 1024. 1000 <= range <= 9999
// retval = volts * volts / (1024 * 1024) * range

uint16_t calcPwr(uint16_t volts, uint16_t range)
{
	uint16_t v = dxMult(volts, volts);	// total shift 10 bit
	v = dxMult(v, range);
	return v;	// (Volts * volts * range) >> 20, that means result 0 to range
}

// send complete set of probe parameters (correction
void sendProbePar(uint8_t rprobe)
{
		uint8_t probechar = rprobe + '0';
		while (UCSRB & (1<<UDRIE));
		txbuf[0] = probechar;
		txbuf[1] = 'G';	// range
		txcount=2;
		uint16_t range = eeprom_read_byte(ranges + rprobe);
		n2ascii(range,0);
		usartNumber();
		txbuf[txcount++] = 13;	//<CR>
		txbuf[txcount++] = 10;	//<LF>
		UCSRB |= (1<<UDRIE);	// send it!!!
		
		while (UCSRB & (1<<UDRIE));
		txbuf[0] = probechar;
		txbuf[1] = 'S';	// scale
		txcount=2;
		uint16_t scale = eeprom_read_word(scales + rprobe);
		n2ascii(scale,0);
		usartNumber();
		txbuf[txcount++] = 13;	//<CR>
		txbuf[txcount++] = 10;	//<LF>
		UCSRB |= (1<<UDRIE);	// send it!!!
		
		for(uint8_t i=0; i<16; i++)
		{
			while (UCSRB & (1<<UDRIE));
			txbuf[0] = probechar;
			txbuf[1] = 'C';	// correction
			txbuf[2] = i + 'A';
			txcount=3;
			uint16_t scale = eeprom_read_byte(corrs[rprobe] + i);
			n2ascii(scale,0);
			usartNumber();
			txbuf[txcount++] = 13;	//<CR>
			txbuf[txcount++] = 10;	//<LF>
			UCSRB |= (1<<UDRIE);	// send it!!!
		}
}


// main program loop. Performs initialisation and goes into infinite loop, performing
// different tasks. The task flags are set by interrupt procedures
int main (void)
{
	initHW();
	uint8_t* pbuf = txbuf;
	*pbuf = 13;
	pbuf++;
	*pbuf = 10;
	pbuf++;
	*pbuf = 'S';
	pbuf++;
	*pbuf = 'W';
	pbuf++;
	*pbuf = 'R';
	pbuf++;
	*pbuf = '1';	// version number (for compatibility check with config program)
	pbuf++;
	*pbuf = '0';
	pbuf++;
	*pbuf = 13;
	pbuf++;
	*pbuf = 10;
	pbuf++;
	txcount = 9;
	UCSRB |= 1<<UDRIE;

	while (1)
	{
		// ADC convertion complete (every 100 msec)
		if (flagreg & (1<<ADCCOMPLETE))	// ADC value set is available
		{
			flagreg &= ~(1<<ADCCOMPLETE);	// reset the flag
			probe = (adcValue[0] >> 7) & 0x07;
			uint16_t fwd = correction(adcValue[1], probe);
			uint16_t rev = correction(adcValue[2], probe);
			fwdavg = (AVERAGEFACTOR * fwdavg + fwd)/(AVERAGEFACTOR + 1);		//floating average
			revavg = (AVERAGEFACTOR * revavg + rev)/(AVERAGEFACTOR + 1);
			memptr++;
			if (memptr >= PEAKMEMSIZE)
				memptr = 0;	// rounding
			fwdmem[memptr] = fwd;	// override oldest values with the new ones
			revmem[memptr] = rev;
			
			uint16_t fwdpeak = fwdmem[0];
			uint16_t revpeak = revmem[0];
			for(uint8_t i = 1; i<PEAKMEMSIZE; i++)
			{
				if (fwdmem[i] > fwdpeak)
					fwdpeak = fwdmem[i];
				if ( revmem[i] > revpeak)
					revpeak = revmem[i];
			}
			if (fwdpeak >=1023)
				flagreg |= (1<<OVERLOAD);	// set overload flag - the forward power is too high!!
			
			
			uint8_t probechar;
			if (PIND & (1<<PD2))
			{
				fwd = fwdpeak;	// display peak values
				rev = revpeak;
				probechar = 'P';	// peak values
			}
			else
			{
				fwd = fwdavg;	// display average values
				rev = revavg;
				probechar = 'A';	// average values
			}
			showBar(fwd, rev);	// show bar indicator

		
			// display refresh (every 500 msec)
			if (flagreg & (1<<REFRESH))	// value processing & display refresh	
			{
				flagreg &= ~(1<<REFRESH);	// reset the flag

				uint16_t swr = 0;	// value not available (unsufficient signal level)
				if (fwdpeak >= 100)		// 10% excitation for FWD signal (1% for nominal tx power)
				{
					swr = 10000;		// SWR very high (>10)
					if (revpeak < fwdpeak)
					{
						swr = ((fwdpeak+revpeak)*20)/(fwdpeak - revpeak);	// calculated value; 20 times real SWR
						if (swr>=2000)
							swr = 10000;	// too high
						else
							swr *= 5; // value = 100 * SWR
					}
				}
			
				uint16_t scale = eeprom_read_word(scales + probe);
				fwd = calcPwr(fwd, scale);	// replace voltage values with power values
				rev = calcPwr(rev, scale);
			
				// show measured values on 2nd display line
				// example for length 40: P2 F=34.56 R=12.76 PWR=22.72kW SWR=3.45
				// example for length 24: P2 P=22.72kW SWR=3.45
				// example for length 20: P2 22.72kW SWR=3.45
				uint8_t range = eeprom_read_byte(ranges + probe);
				lcdSendCmd(0xc0);	// we start at 1st position on 2nd line
				lcdSendChar(probechar);
				lcdSendChar(probe + '0');
#if LCDLINELENGTH==40 
				lcdEText(txtF);	// forward
				n2ascii(fwd, range & 0x0F);	// lower bits - decimal point position
				lcdNumber();
				lcdEText(txtR);
				n2ascii(rev, range & 0x0F);	// lower bits - decimal point position
				lcdNumber();
				lcdEText(txtPwr);
#endif
#if LCDLINELENGTH==20
				lcdSendChar(' ');
#endif
#if LCDLINELENGTH==24
				lcdEText(txtP);
#endif
				if (fwd > rev)
					fwd = fwd - rev;	// forward power minus reverse power
				else
					fwd = 0;
				n2ascii(fwd, range & 0x0F);	// lower bits - decimal point position
				lcdNumber();
				if (!(flagreg & (1<<SETUPMODE)))
				{
					while (UCSRB & (1<<UDRIE));
					txbuf[0] = 'P';	// power
					txbuf[1] = 'w';
					txcount=2;
					usartNumber();
					UCSRB |= (1<<UDRIE);	// send it!!!
				}
				uint8_t pfx = ' ';
				switch(range >> 4)
				{
					case 0: pfx='p';break;
					case 1: pfx='n';break;
					case 2: pfx=0xe4;break; // micro
					case 3: pfx='m';break;
					case 5: pfx='k';break;
					case 6: pfx='M';break;
				}
				lcdSendChar(pfx);
				if (!(flagreg & (1<<SETUPMODE)))
				{
					while (UCSRB & (1<<UDRIE));
					txbuf[0] = pfx;	// power
					txbuf[1] = 'W';
					txbuf[2] = 13;	// <CR>
					txbuf[3] = 10;	// <LF>
					txcount=4;
					UCSRB |= (1<<UDRIE);	// send it!!!
				}

				lcdEText(txtSWR);
				if (swr == 0)
				{
					lcdEText(txtUnknown);	// power too low, no calculation
				}
				else if (flagreg & (1<<OVERLOAD))
				{
					lcdEText(txtOverload);	// FWD signal overload, calculation wrong
					flagreg &= ~(1<<OVERLOAD);	// reset the overload flag
				}
				else
				{
					if (swr <1000)
					{
						n2ascii(swr,2);
						lcdNumber();	// calculated value (100 times the value, range 100 - 999)
						if (!(flagreg & (1<<SETUPMODE)))
						{
							while (UCSRB & (1<<UDRIE));
							txbuf[0] = 'S';	// power
							txbuf[1] = 'W';
							txbuf[2] = 'R';
							txcount=3;
							usartNumber();
							txbuf[txcount++] = 13;	//<CR>
							txbuf[txcount++] = 10;	//<LF>
							UCSRB |= (1<<UDRIE);	// send it!!!
						}
					}
					else	
						lcdEText(txtHigh);	// SWR too high...
				}
			
			
				if (flagreg&(1<<SETUPMODE))
				{
					setupUsartSend();	// periodic sending of ADC values in setup mode
				}
			}
		}
		
		// command received over serial port
		if (flagreg & (1<<RXCOMPLETE))		// instruction received
		{
			// "set" - enter setup mode
			if ((rxcount==3)&&(rxbuf[0] == 's')&&(rxbuf[1]=='e')&&(rxbuf[2]=='t'))
			{
				flagreg |= (1<<SETUPMODE);
			}

			
			// setup mode commands (possible in setup mode only)
			if (flagreg & (1<<SETUPMODE))
			{
				// "q" - quit setup mode
				if ((rxcount==1)&&(rxbuf[0]=='q'))
				{
					flagreg &= ~(1<<SETUPMODE);
				}
				else if (rxbuf[0] == 'w')
				{
					// "wp" - EEPROM write; example wp1c123 : write point 2 ('c') of probe 1 ('1') value 123 ('123') 
					if ((rxcount>4) && (rxbuf[1] == 'p'))
					{
						uint8_t wprobe = rxbuf[2] - '0';
						if (wprobe < PROBES)
						{
							uint8_t wpoint = rxbuf[3] - 'a';
							if (wpoint < 16)
							{
								uint16_t val = ascii2n(rxbuf + 4, rxcount - 4);
								eeprom_write_byte(corrs[wprobe]+wpoint, val);
								while ((UCSRB & (1<<UDRIE)));	// wait till usart ready (UDRIE cleared by interrupt rutine)
								txbuf[0] = 'W';
								txbuf[1] = 'P';
								txbuf[2] = 13;	//CR
								txbuf[3] = 10;	//LF
								txcount = 4;
								UCSRB |= (1<<UDRIE);	// send confirmation ("OK");
							}
						}
					}
					
					// "ws" - EEPROM write; example ws11235 : write scale for probe 1 ('1') value 1235 ('1235') 
					if ((rxcount>6) && (rxbuf[1] == 's'))
					{
						uint8_t wprobe = rxbuf[2] - '0';
						if (wprobe < PROBES)
						{
							uint16_t val = ascii2n(rxbuf + 3, rxcount - 3);
							if ((val < 9999) & (val >= 1000))
							{
								
								eeprom_write_word(scales + wprobe, val);
								while ((UCSRB & (1<<UDRIE)));	// wait till usart ready (UDRIE cleared by interrupt rutine)
								txbuf[0] = 'W';
								txbuf[1] = 'S';
								txbuf[2] = 13;	//CR
								txbuf[3] = 10;	//LF
								txcount = 4;
								UCSRB |= (1<<UDRIE);	// send confirmation ("OK");
							}
						}
					}
					
					// "wr" - EEPROM write; example wr1222 : write range for probe 1 ('1') value 222 ('222') 
					if ((rxcount>3) && (rxbuf[1] == 'r'))
					{
						uint8_t wprobe = rxbuf[2] - '0';
						if (wprobe < PROBES)
						{
							uint16_t val = ascii2n(rxbuf + 3, rxcount - 3);
							if (val < 256) 
							{
								
								eeprom_write_byte(ranges + wprobe, val);
								while ((UCSRB & (1<<UDRIE)));	// wait till usart ready (UDRIE cleared by interrupt rutine)
								txbuf[0] = 'W';
								txbuf[1] = 'R';
								txbuf[2] = 13;	//CR
								txbuf[3] = 10;	//LF
								txcount = 4;
								UCSRB |= (1<<UDRIE);	// send confirmation ("OK");
							}
						}
					}
				}
				
				// "r3" - read all EEPROM parameters for probe 3
				else if ((rxcount==2) && (rxbuf[0] == 'r'))
				{
					uint8_t rprobe = rxbuf[1] - '0';
					if ((rprobe >= 0) && (rprobe < 8))
						sendProbePar(rprobe); // ToDo send all probe parameters via USART
				}
			}
			rxcount = 0;
			flagreg &= ~(1<<RXCOMPLETE);	// reset the flag
		}
	}
}



